/******************************************************************************
turtle: turlegrafik

Copyright (C) 2011	Günther Schardinger,
                    Forstweg 12,
                    8793-Trofaiach, Austria
                    g.schardinger@aon.at

Dieses Programm ist freie Software. Sie können es unter den Bedingungen der
GNU General Public License, wie von der Free Software Foundation veröffentlicht,
weitergeben und/oder modifizieren, entweder gemäß Version 3 der Lizenz oder
jeder späteren Version.

Die Veröffentlichung dieses Programms erfolgt in der Hoffnung, daß es Ihnen von
Nutzen sein wird, aber OHNE IRGENDEINE GARANTIE, sogar ohne die implizite
Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK.
Details finden Sie in der GNU General Public License.
GNU General Public License siehe <http://www.gnu.org/licenses/>.
*******************************************************************************/
#ifndef GRAFIK_H
#define GRAFIK_H
#include <QtGui>
#include <QtScript>
#include "utils/utils.h"
#include "_turtle.h"


/*! Grafikmodule >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
Die Verwendung der Grafik Klassen:

  Dockfenster:  tGDock
  Widget:       tGrafik, tGrafik_Info, tTurtle
  Verwaltung:   tGalerie, tGalerieDialog und tSteuerung

Die Klasse tGDock stellt Dock-Fenster zum Anzeigen von tGrafik-Widgets
zur Verfuegung. Die von der Basis-Klasse tGrafik abgleiteten eigenen
Grafik-Klassen koennen Grafiken berechnen und ohne Änderung auch
als PDF-Datei speichern.

Für die Grafikberechnung stellt die Komfortklasse tTurtle neben dem Painter
auch ein mathematisches Koordinatensystem und weitere Zeichenbefehle zur
Verfügung.

Zur Auswahl der verschiedenen Grafiken (Grafikfenster) dient die Klasse
tGalerie. Der tGalerieDialog bietet Dialogfenster fuer die manuellen Auswahl.
tGalerie kann Grafiken aber auch ueber externe Scripts steuern.

-------------------------------------------------------------------------------
Die Basisklassen fuer alle Grafiken sind: tGrafik_Info und tGrafik.

class tGrafik_Info:
     Wird von tGalerie zum Anzeigen der Grafik-Titel und
     zum Erzeugen der speziellen Grafikinstanzen verwendet.
     * Titel() Eindeutiger Titel der Grafik.
     * Type() Grafiktype
     * New() Erzeugt die spezielle Grafik mit new().

class tGrafik:
     Basis-Widget zum Zeichnen von Grafiken am Bildschirm oder in PDF-Datei.
     Die Zeichenbefehle fuer die Grafik werden in der Funktion draw() programmiert.
     Ausgefuehrt werden diese Befehle von der Klasse tTurtle.

     Features von draw() bereitgestellt von tTurtle:
     * DIN Papierformate, Zeichen-Einheiten in mm (qreal)
     * Koordinatensystem: x-Achse nach rechts, y-Achse nach oben, Ursprung
     * Zum Zeichen steht ein initialisierter Painter zur Verfuegung
     * Komfortklasse tTurtle liefert vereinfachte Zeichenbefehle
     * In draw() koennen auch Scriptbefehle ausgefuehrt werden.

     Siehe auch class tTurtle.

-------------------------------------------------------------------------------
Eine neue Grafik 'tMyGrafik' erstellen: Beispiele siehe Ordner grafiken/.

     1. Neue Klasse 'tMyGrafik' von tGrafik ableiten.
        Die Funktionen Anleitung() und draw() implementieren.

        Sollte die Grafik auch durch Scripts gesteuert werden:
           Unter Anleitung() ein Scriptbeispiel ablegen.

        Sollte die Grafik Setup-Informationen speichern:
           Die Setup-Funktionen save(tMyGrafik *) und
           restore(tMyGrafik *) implementieren.

     2. Neue Klasse 'tMyGrafik_Info' von tGrafik_Info ableiten.
        Titel() und New() implementieren.

     3. Die neue Klasse 'tMyGrafik_Info' im Konstruktor von tGalerie()
        eintragen. Damit scheint der Titel() im GalerieDialog auf
        und die Grafik kann manuelle oder durch Scripts
        instantiert/angezeigt werden.

<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/


// class tGrafik >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
// >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

class tGrafik : public tTurtle   // Basisklasse
{ Q_OBJECT
 public:
   tGrafik(QWidget *parent=0);
   virtual ~tGrafik();
                                  // Grafiktypen
   enum GType{  GCTool   =0x01,   // C++ Grafiktool, löscht GDock nicht!
                GCPlus   =0x02,   // C++ Grafik
                GCTest   =0x04,   // C++ Testgrafik
                GScript  =0x08,   // C++ Grafik für Scripts
                GNoType  =0x00};  // undefiniert


   virtual QString Anleitung();   // Anleitung zur Grafik
   virtual void    draw();        // Zeichnenbefehle fuer die Grafik

   virtual void save(QSettings *) {}    // Setup Daten sichern
   virtual void restore(QSettings *) {} // Setup Daten laden

   // Scripte =================================================================
   //  Zum Auswerten von Scripten kann tGrafik eine
   //  ScriptEngine übernehmen und verwalten.

   // Die Verwendung von Scripten in tMeineGrafik:
   //
   //  1. Variable QScriptValue thisScript; deklarieren
   //
   //  2. tMeineGrafik::draw()
   //      { ...
   //        ... getScriptVar(...)
   //        ...
   //        initScript(thisScript);
   //        ...
   //        if (!Begin());
   //        ...
   //        drawScript(thisScript);
   //        ...
   //       }

   QString getScriptVar(const QString &VarName, const QString &Default);
   int     getScriptVar(const QString &VarName, int Default, int min, int max);
   qreal   getScriptVar(const QString &VarName, qreal Default);
   // Script-Variablen abfragen und Fehlerfall Defaultwerte setzen
   // Verwendung in irgendwo in tMeineGrafik::draw()

   bool initScript(QScriptValue &thisScript);
   // * initScript() muss vor tTurtle.Begin() aufgerufen werden.
   // * Es wird zuerst tTurtle::init() aufgerufen und dann die
   //   Scriptfunktion 'function init(){ ... }' ausgewertet.
   // * thisScript wird mit tMeineGrafik intitialisiert.
   //   Alle 'public slots:' von tMeineGrafik können im Script mit
   //   Object "t" aufgerufen werden.  Beispiel: t.msg("Meldung");

   bool drawScrip(const QScriptValue &thisScript);
   // * drawScrip() kann nach tTurtle.Begin() aufgerufen.
   // * Führt die Scriptfunktion 'function draw()' aus.
   //   In 'function draw()' kann auch gezeichnet werden.
   //   Verwendung nach initScript(thisScript)

   bool hasScript() {return ScriptEngine;}
   // tMeineGrafik besitzt ScriptEngine, also ein Script.

   bool callScript(const QScriptValue &thisScript, const QString &Bezeichner);
   // * Scriptfunktion 'function Bezeichner()' auswerten.
   //   Verwendung nach initScript(thisScript)

   bool callBefehl(const QString &BefehlsCode, const QString &Befehl );
   // String mit Befehlscode auswerten:
   // BefehlsCode="function test() { t.move(40;}"
   // Befehl="test"

   static bool hasErr(QScriptEngine *engine, bool show=true);
   // Rueckgabe: true  -> Scriptfehler, bei show==true Fehlermeldung anzeigen
   //            false -> ok

   QString ScriptSource(const QString Bezeichner);
   // Source-Code von Variablen oder Funktionen anzeigen

   QScriptEngine *Script(){return ScriptEngine;}
   // ScriptEngine verwenden

   void setEngine(QScriptEngine *ScriptEngine);
   // Eine neue ScriptEngine als Besitzer übernehmen.
   // Die alte ScriptEngine wird deleted.
   // Der Destruktor von tGrafik deletes ScriptEngine

   void delEngine();
   // delete ScriptEngine

signals:
   void markScriptLine(int Zeile); // Scriptzeile im Textdock markieren
                                   // für Fehleranzeige

   // Ende Scripte ============================================================


public slots:
   /*! Begin class tGrafik Javascript  **********************************/
   /*! |t.Grafik|********************************************************/
       //Erweiterte Turtlebefehle

   /*! |Koordinatensystem|***********************************************/
       //Koordinatensystem zeichnen

   void drawKoordSystem(bool Beschriftung=true, qreal Einheit=10, qreal PSize=0.6);
        //Beschriftung true/false.
        //Einheit: Abstand der Marken in mm.
        //PSize: Grösse in mm.

   void drawRaster(qreal Einheiten_mm=10, qreal PSize=0.6);
        //Rastermarken zeichnen.
        //Einheit: Abstand in mm.
        //PSize: Grösse in mm.

   /*! |Meldungen|*******************************************************/
       //Textausgaben im Meldungsfenster

   void msg(QVariant s=" ", const QString rgb="black");
        //Meldung s im Meldungsfenster ausgeben
   void clrMsg();
        //Meldungsfenster löschen

   void msgScript(bool showWerte=true);
        //Javascript-Objekte des Scripts in der Msgbox ausgeben.
        //true : Objektnamen und Werte.
        //false: Nur Objektnamen.

   QString infoScript(bool showWerte=true);
           //String für msgScript()


   /*! |Farben und Zufall|***********************************************/

   qreal   rnd(qreal max);
           //Zufallszahlen aus [0,max]
   qreal   rndW(qreal min, qreal max);
           //Zufallszahlen [min,max] und [-min,-max]

   void    setRndRGB(const QString &RGB1="", const QString &RGB2="black");
           //Anfangsfarben für rndRGB(), rndRGB1() und rndRGB2()
   void    setRndHeller(int Heller=100, int rndHeller=0);
           //Aufhellung l für rndRGB(), rndRGB1() und rndRGB2().
           //l=Heller + rnd(rndHeller).
           //0<l bis l<100: Ergebnis dunkler.
           //100<l bis l<400: Ergebnis heller.

   QString rndRGB(qreal rnd=0);
           //Farbe zwischen RGB1 und RGB2. 0<=rnd<=1.

   QString rndRGB1();
           //Zufallsfarbe zu RGB1
   QString rndRGB2();
           //Zufallsfarbe zwischen RGB1 und RGB2

   QString rgbName(int rot, int gruen, int blau);
           //Rückgabe RGB-String

   /*! |Diashow|*********************************************************/

   qreal showDia(qreal ms, qreal Startwert, qreal Decrement=1);
         //Grafik alle ms neu zeichnen.
         //Startwert mit Decrement herunterzählen.
         //Rückgabe: DiaRestwert in ms.
         //
         //if (DiaRestwert<=0)	DiaRestwert = Startwert - Decrement;
         //else	DiaRestwert = DiaRestwert - Decrement;

         //if (DiaRestwert>0) 	Neuzeichnen in ms.
         //else	Stop

   /*! |t.Befehle|*******************************************************/
       //Eigene Befehle für Funktionen und Polygonobjekte.

   /*! |Befehle: Funktionen|*********************************************/
       //Funktionen werden in der Form
       //function 'Name'() { Befehl; Befehl; ... }
       //gespeichert.

   void defCall(const QString Name, const QString Befehle, int mode=0);
        //Funktionsdefinition 'Name' anlegen oder erweitern.
        //mode=0: Funktion 'Name' neu anlegen und Befehle anhängen.
        //mode=1: Befehle an Funktion "Name" anhängen.

   void recDraw(QString Name="");
        //Befehlsrecorder. Turtlebefehle als Funktion aufzeichnen.
        //Name<>"": Funktion 'Name' anlegen und Aufzeichnugsmodus ein.
        //Alle folgenden Turtlebefehle werden aufgezeichnet.
        //Name=="": Aufzeichnung beenden.

   void call(const QString Name);
        //Funktion 'Name' ausführen.
        //Turtlestatus nicht speichern.

   /*! |Befehle: Polygone definieren|************************************/
       //Mit def-Aufrufen werden verschachtelte
       //Polygonobjekte definiert.

   void defPoint(const QString Name);
        //Punkt (lastY(),lastY()) als Punkt "Name" speichern

   void defPoint(const QString Name, int Index);
        //Punkt aus Turtlepolygon kopieren.
        //Polygon[Index] unter "Name" speichern.
        //Index=-1 letzter Punkt.

   void defPoint(const QString Name, const QString P1, const QString P2, qreal t);
        //Teilungspunkt P = P1 + t*(P2-P1) als Punkt "Name" speichern

   void defVector(const QString Name, const QString P1, const QString P2, qreal t=1);
        //Ortsvektor t*(P2-P1) als Punkt "Name" speichern

   void defLine(const QString Name,  qreal l);
        //P1 (lastY(),lastY()) und P2 move(l) mit LastW()
        //P1 und P2 als Befehls-Polygon "Name" speichern

   void defLine(const QString Name, const QString P1, const QString P2);
        //Punkte P1 und P1 als Befehls-Polygon "Name" speichern

   void defDraw(const QString Name, int mode=0);
        //Speichert unter 'Name' Listen von Polygonobjekten.
        //Ein Polygonobjekt übernimmt von der Turtle die
        //Daten: Polygon, Pen, Brush, FillMode.
        //
        //mode=0: 'Name' neu anlegen und Turtledaten speichern.
        //mode=1: Turtledaten an Befehl 'Name' anhängen.
        //mode=2: Polygon zuerst am ersten Polygonobjekt
        //klippen und  Turtledaten anhängen.
        //mode=3: Einpassen:
        //Das erste Polygon [N0, .. Nm] von 'Name'
        //durch das Turtlepolygon [P0, .. Pn] ersetzen.
        //N0!=Nm: Turtlepolygon in Polygon 'Name'
        // einpassen. P0 -> N0 und Pn -> Nm
        //N0==Nm: Turtlepolygon verschieben mit P0 -> N0


   bool defScript(const QString Flag, const QString Grafikname);
        //Befehle eines anderen Scripts in die
        //Befehlsliste eintragen und ein Funktions-Flag setzen.
        //
        //Funktion Flag() ist in Befehlsliste:
        //  * nichts tun
        //  * Rückgabe true
        //Flag ist nicht in Befehlsliste:
        //  * Script 'Grafikname' ausführen
        //  * Funktion Flag() eintragen
        //  * Aktuelles Script neu starten
        //  * Rückgabe false

   /*! |Befehle: Polygone zeichnen|**************************************/
       //Alle Polygonobjekte werden an der aktuellen Turtleposition,
       //mit der aktuellen Turtlerichtung und unter
       //Verwendung der Abbildung 'drawMatrix' ausgeführt.

   void draw(const QString Name, bool UseStyle=true);        
        //'Name' ist Polygonobjekt:
        // Befehl an der aktuellen Turtleposition
        // und mit Turtlerichtung und unter
        // Verwendung der drawMatrix ausführen.
        //'Name' ist Funktion: Funktion ausführen.
        // UseStyle==true : Definierten Style verwenden
        // UseStyle==false: Aktuellen Turtle-Style verwenden
        //Turtlestatus bleibt erhalten.

   void drawName(const QString Name, qreal dx=0, qreal dy=0);
        //Beschriftung 'Name' an Objektposition anzeigen.
        //Position um (dy,dy) veschieben.

   void shift(const QString Name,  qreal t=1);
        //Ein moveTo auf: (lastX(),lastY()) + t*("Name"))
        //Drehung: um lastW();



   /*! |Befehle: Abbildungen|********************************************/
       //Befehle zum Erzeugen der Abbildungsmatrix 'drawMatrix'
       //Achtung: Die Einstellungen verändern die bestehende
       //Matrix kumulativ und sind nicht kommutativ!
       //reset=true : Abbildung auf die Einheitsmatrix anwenden.
       //reset=false : Abbildung auf drawMatrix anwenden.

   void setRotate(qreal Winkel, bool reset=true);
        //Drehung um Punkt(0,0), Winkel in Grad.
   void setRotate(const QString Punkt, qreal Winkel, bool reset=true);
        //Drehung um Punkt, Winkel in Grad.
   void setMirror(const QString Punkt1, const QString Punkt2, bool reset=true);
   void setMirror(qreal x0, qreal y0, qreal x1, qreal y1, bool reset=true);
        //Spiegelung an der Geraden (Punkt1,Punkt2);
   void setScale(qreal sx, qreal sy, bool reset=true);
   void setShear(qreal sh, qreal sv, bool reset=true);
   void setTranslate(qreal dx, qreal dy, bool reset=true);
        //Schiebung um (dy,dy)
   void setIdentity();
        //Einheitsmatrix (reset)


   /*! |Befehle: Turtlepolygon|**************************************************/
       //Befehle zum Abbilden des Turtlepolygons
       //mit 'drawMatrix'
       //Der erste Polygonpunkt hat den Index 0!

   void setPolygon(int Index=0, int Anzahl=-1);
        //Turtlepolygons durch eine Abbildung mit drawMatrix ersetzen.
        //Der erste Polygonpunkt hat den Index 0!
        //Der letzte Polygonpunkt hat den Index -1.
        //
        //Index= 0, Anzahl=-1: Alle Punkte verwenden.
        //Index=-1, Anzahl=-1: Alle Punkte, Reihenfolge umgekehrt.
        //Index>=0, Anzahl>=0: Anzahl Punkte ab Index verwenden.
        //Index<-1, Anzahl>=0: Anzahl Punkte ab Index, Reihenfolge umgekehrt.


   void addPolygon(const QString Name, int mode=0);
        //Turtlepolygon erweitern.
        //Das erste Polygon von "Name" mit drawMatrix abbilden
        //und zum Endpunkt des Turtlepolygons verschieben.
        //mode=0 : Punkte in gegebener Reihenfolge anhängen.
        //mode=-1: Punkte in umgekehrter Reihenfolge anhängen.

   void setPolygon(const QString Name, int mode=0, qreal dt=0.1);
        //Turtlepolygon aus dem ersten Polygon von "Name" erzeugen.
        //"Name" wird zuerst mit drawMatrix abgebildet.
        //mode=0: Turtlepolygon durch "Name" ersetzen.
        //mode=-1: Turtlepolygon durch "Name" ersetzen. Reihenfolge umgekehrt.
        //mode=1: Turtlepolygon Duchschnitt "Name".
        //mode=2: Turtlepolygon Vereinigung "Name".
        //mode=3: Turtlepolygon geschlossen: Turtlepolygon ohne "Name".
        //Turtlepolygon offen: Turtlepolygon mit "Name" abschneiden.
        //mode = 4: "Name" liefert die Stützpunkte eines quadratischen Bezier Splines.
        //epsilon<dt<1 Bestimmt die Punktedichte pro Spline-Segment.

   void  msgPolygon(bool Spline=false);
         //Polygon als Funktion im Meldungsfenster anzeigen.
         //Kann ins direkt ins Script kopiert werden
         //Spline==true: Splinepolygon erzeugen


   /*! |Befehle: Sonstiges|************************************************/
       // Verwaltung

   bool exists(const QString Name);
        //Rückgabe true: Befehl 'Name' existiert.

   void msgDraw(const QString Name="");
        //'Name'=="": Alle Befehle der Befehlsliste anzeigen
        //'Name'!="": Infos zum Befehl 'Name'

   QString drawInfo(const QString Name);
        //msgDraw() als String.

   void clrDraw(const QString &Name);
        //Befehl 'Name' löschen.

   void clrDraw();
        //Befehlsliste löschen

   /*! End Javascript **************************************************/

protected:
  void setMirror(QPointF A, QPointF B, bool reset);

  //void resizeEvent ( QResizeEvent * event );
  //void mouseDoubleClickEvent(QMouseEvent * event);

private:
   QScriptEngine *ScriptEngine;   // 0 oder Scriptbefehle fuer draw()
                                  // tGrafik übernimmt die Scriptengine
                                  // der Destruktor löscht das Objekt!

   qreal          DiaRestwert;    // Zählerstand für showDia



   friend class   tDrawBefehl;    // draw Befehle
   QMatrix        drawMatrix;     // Matrix für draw(Befehl)

};


// ---------------------------------------------------------------------------
// tGrafik_Info legt Titel und Type der Grafik fest und
// kann das Objekt mit New() erzeigen.
//
class tGrafik_Info
{ public:
    virtual QString  Titel()      {return "No Show";}        // Titel der Grafik
    virtual tGrafik::GType Type() {return tGrafik::GCPlus ;} // Grafiktype

    virtual tGrafik *New(QWidget *parent=0) {return new tGrafik(parent);}
    virtual ~tGrafik_Info(){}
};
// Ende tGrafik <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


#endif // GRAFIK_H
